from typing import NewType

TaskID = NewType("TaskID", str)
DedupKey = NewType("DedupKey", str)
UserID = NewType("UserID", str)
CompanyID = NewType("CompanyID", str)
RecordingID = NewType("RecordingID", str)
